#!/usr/bin/env python3
import os
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import linear_sum_assignment
import seaborn as sns
sns.set_style('ticks')
sns.set_context('talk', font_scale=0.7)

def get_test():
  # Rainer E. Burkard and Ulrich Derigs
  # Assignment and Matcing Problems: Solution Methods
  #  with FORTRAN_Programs
  # Springer (1980). p. 11, cost 76
  cmat = np.array([
    [13, 21, 20, 12, 8, 26, 22, 11],
    [12, 36, 25, 41, 40, 11, 4, 8],
    [35, 32, 13, 36, 26, 21, 13, 37],
    [34, 54, 7, 8, 12, 22, 11, 40],
    [21, 6, 45, 18, 24, 34, 12, 48],
    [42, 19, 39, 15, 14, 16, 28, 46],
    [16, 34, 38, 3, 34, 40, 22, 24],
    [26, 20, 5, 17, 45, 31, 37, 43],
  ])
  x = np.array([1, 8, 7, 5, 2, 6, 4, 3]) - 1
  return cmat, x

def get_test1():
  # Rainer Burkard, Mauro Dell'Amico, and Silvano Martello
  # Assignment Problems: Revised Reprint
  # SIAM (2009). Chap. 4, p. 77
  cmat = np.array([
    [7, 9, 8, 9],
    [2, 8, 5, 7],
    [1, 6, 6, 9],
    [3, 6, 2, 2]
  ])
  cbar = np.array([
    [0, 0, 1, 2],
    [0, 4, 3, 5],
    [0, 3, 5, 8],
    [1, 2, 0, 0]
  ])
  # row -> column
  phi = np.array([0, -1, -1, 2])
  # column -> row
  row = np.array([0, -1, 3, -1])
  return cmat, cbar, phi, row

def dual_variables(cmat):
  u = np.min(cmat, axis=1)
  v = np.min(cmat-u[:, None], axis=0)
  return u, v

def admissible_transform(cmat, u, v):
  m, n = cmat.shape
  assert m == n
  cbar = np.array([
    [cmat[i, j]-u[i]-v[j] for j in range(n)]
  for i in range(n)])
  return cbar

def partial_primal_solution(cmat, ztol=1e-8):
  m, n = cmat.shape
  assert m == n
  u, v = dual_variables(cmat)
  cbar = admissible_transform(cmat, u, v)
  sel = abs(cbar) < ztol
  nvec = np.arange(n)
  phi = -np.ones(n, dtype=int)  # null assignment
  y0 = np.zeros(n)  # null cost
  z0 = -np.ones(n, dtype=int)  # inverse assignment
  icol = np.zeros(n, dtype=bool)
  irow = 0
  for rsel, row in zip(sel, cmat):
    j = np.argmin(row[rsel])
    k = nvec[rsel][j]  # candidate
    if not icol[k]:
      phi[irow] = k
      y0[k] = row[k]
      z0[k] = irow
      icol[k] = True
    else:
      if row[k] <= y0[k]:
        phi[irow] = k
        y0[k] = row[k]
        phi[z0[k]] = -1  # unassign previous
        z0[k] = irow
    irow += 1
  return phi

def transpose_map(phi):
  n = len(phi)
  row = -np.ones(n, dtype=int)
  for i, j in enumerate(phi):
    if j >= 0:
      row[j] = i
  return row

def full_set(n):
  return set(range(n))

def minimum_reduced_cost(cbar, su, lv):
  m, n = cbar.shape
  assert m == n
  vset = full_set(n)
  delta = np.inf
  for i in su:
    for j in vset-lv:
      if cbar[i, j] < delta:
        delta = cbar[i, j]
  return delta

def alternate(k, row, cbar, verbose=False, return_path=False):
  n = len(row)
  jpred = -np.ones(n, dtype=int)
  su = set()  # selected rows
  lv = set()  # labeled columns
  sv = set()  # selected columns
  vset = full_set(n)
  i = k
  fail = False
  sink = -1
  if return_path:
    path = []
  iwhile = 0
  while (not fail) and (sink < 0):
    if verbose:
      print(iwhile, sink, fail)
    su.add(i)
    for j in vset-lv:  # look for admissible column at row i
      if (cbar[i, j] == 0):  # admissible
        lv.add(j)
        jpred[j] = i
    # scan lv by selecting one column at a time
    if len(lv-sv) > 0:  # scan next column in lv
      j = (lv-sv).pop()
      sv.add(j)
      if row[j] < 0:
        sink = j
        if return_path:
          path.append((jpred[j], j))
      else:
        #if return_path:
        #  path.append((jpred[j], j))
        if return_path:
          path.append((i, j))
        i = row[j]
        if return_path:
          path.append((i, j))
          #path.append((jpred[j], j))
    else:  # finished scan without finding sink
      fail = True
    iwhile += 1
  if verbose:
    print(iwhile, sink, fail)
  ret = (sink, jpred, su, lv, sv)
  if return_path:
    ret = (sink, jpred, su, lv, sv, path)
  return ret

def alternating_path(k, jpred, sink, row0, phi0):
  assert sink >= 0
  row = row0.copy()
  phi = phi0.copy()
  j = sink
  i = -1
  ijl = []
  while i != k:
    ijl.append(j)
    i = jpred[j]
    row[j] = i
    ijl.append(i)
    phi[i], j = j, phi[i]
  #print(sink)
  #print(ijl)
  #print(k)
  path = []
  #path.append((ijl[0], -1))
  for i, j in zip(ijl[:-1], ijl[1:]):
    #path.append((i, j))
    path.append((j, i))
  #path.append((-1, ijl[-1]))
  print('alt', path)
  return path

def cost(cmat, phi):
  sel = phi >= 0
  z = 0
  for row, j in zip(cmat[sel], phi[sel]):
    z += row[j]
  return z

def show_cost_matrix(ax, cmat):
  m, n = cmat.shape
  assert m == n
  x = np.array(n*[np.arange(n)], dtype=int).ravel()
  y = np.array([[irow]*n for irow in range(n)], dtype=int).ravel()
  ax.matshow(cmat)
  ax.get_xaxis().tick_bottom()
  ## !!!! y should start from the top
  #yticks = range(n)
  #ax.set_yticks(yticks)
  #ax.set_yticklabels(yticks[::-1])
  for x1, y1, label in zip(x, y,cmat.ravel().astype(str)): 
    ax.text(x1, y1, label, ha='center', va='center',
      c='white', fontsize=36)

def show_assignment(ax, phi, ms=36):
  sel = phi >= 0
  y = np.where(sel)[0]
  x = phi[sel]
  ax.plot(x, y, ls='', marker='o', fillstyle='none',
    ms=ms, c='r', mew=5,
  )

def show_path(ax, path):
  x, y = np.array(path).T
  ax.plot(y, x, lw=8)

def plot_solution(cmat, cbar, phi, path=None, i0=None, j0=None,
  fig_loc = None):
  fig, axl = plt.subplots(1, 2, sharex=True, sharey=True)
  axl[0].set_xlabel('cost')
  axl[1].set_xlabel('reduced cost')
  show_cost_matrix(axl[0], cmat)
  show_cost_matrix(axl[1], cbar)
  for ax in axl:
    show_assignment(ax, phi)
    if path is not None:
      axl[0].set_title('row %d with col %d' % (i0, j0))
      show_path(ax, path)
  fig.tight_layout()
  fig.subplots_adjust(hspace=0.025)
  if fig_loc is not None:
    fig.savefig(fig_loc, dpi=320)
  plt.show()

def kuhn_hungarian(cmat, mwhile=1024, verbose=False, visualize=False,
  fig_dir='figures'):
  m, n = cmat.shape
  assert m == n
  u, v = dual_variables(cmat)
  cbar = admissible_transform(cmat, u, v)
  # use my initial
  phi = partial_primal_solution(cmat)
  row = transpose_map(phi)
  if visualize:
    fig_dir = 'n%d-s%d' % (n, iseed)
    isol = 0
  # main loop
  iwhile = 0
  uset = set(np.where(phi>=0)[0])
  while (len(uset) < n) and (iwhile < mwhile):
    k = (full_set(n)-uset).pop()
    # find alternating path
    sink, jpred, su, lv, sv, path = alternate(k, row, cbar,
      return_path=True)
    if visualize:
      if sink >= 0:
        pass
        #path = alternating_path(k, jpred, sink, row, phi)
      else:
        path = None
      fig_loc = os.path.join(fig_dir, 's%03d.png' % isol)
      plot_solution(cmat, cbar, phi, path=path, i0=k, j0=sink,
        fig_loc=fig_loc)
      isol += 1
    if verbose:
      print(iwhile, sink, jpred)
    if sink >= 0:  # enlarge primal solution
      if visualize:
        row0 = row.copy()
        phi0 = phi.copy()
      uset.add(k)
      j = sink
      i = -1
      while i != k:
        i = jpred[j]
        row[j] = i
        phi[i], j = j, phi[i]
    else:  # update the dual solution
      delta = minimum_reduced_cost(cbar, su, lv)
      for i in su:
        u[i] += delta
      for j in lv:
        v[j] -= delta
      cbar = admissible_transform(cmat, u, v)
    if visualize:
      if sink >= 0:
        pass
        #path = alternating_path(k, jpred, sink, row, phi)
      else:
        path = None
      fig_loc = os.path.join(fig_dir, 's%03d.png' % isol)
      plot_solution(cmat, cbar, phi, path=path, i0=k, j0=sink,
        fig_loc=fig_loc)
      isol += 1
    uset = set(np.where(phi>=0)[0])
    iwhile += 1
  #if visualize:
  #  if sink >= 0:
  #    pass
  #    #path = alternating_path(k, jpred, sink, row0, phi0)
  #  else:
  #    path = None
  #  plot_solution(cmat, cbar, phi, path=path)
  return phi

if __name__ == '__main__':
  #cmat, x0 = get_test()

  #cmat, cbar0, phi0, row0 = get_test1()

  ## use default initial
  #phi = phi0
  #row = row0

  cmat = np.array([
    [64, 55, 70, 61],
    [54, 62, 70, 62],
    [92, 95, 87, 99],
    [82, 85, 77, 89],
  ])

  n = 4
  iseed = 7
  iseed = 1836
  iseed = 137
  np.random.seed(iseed)
  cmat = np.random.randint(1, 10, (n, n))
  # ============= end problem definition =============

  fig_dir = 'n%d-s%d' % (n, iseed)
  visualize = True
  if visualize and (not os.path.isdir(fig_dir)):
    os.mkdir(fig_dir)
  phi = kuhn_hungarian(cmat, visualize=visualize, fig_dir=fig_dir)
  phi1 = linear_sum_assignment(cmat)[1]
  if np.allclose(phi, phi1):
    print('same as scipy')
  else:
    print(phi, cost(cmat, phi))
    print(phi1, cost(cmat, phi1))
# end __main__
