#!/usr/bin/env python3
import os
import sys
import numpy as np
from timeit import default_timer as timer

def dual_variables(cmat):
  u = np.min(cmat, axis=1)
  v = np.min(cmat-u[:, None], axis=0)
  return u, v

def admissible_transform(cmat, u, v):
  m, n = cmat.shape
  assert m == n
  cbar = np.array([
    [cmat[i, j]-u[i]-v[j] for j in range(n)]
  for i in range(n)])
  return cbar

def partial_primal_solution(cmat, ztol=1e-8):
  m, n = cmat.shape
  assert m == n
  u, v = dual_variables(cmat)
  cbar = admissible_transform(cmat, u, v)
  sel = abs(cbar) < ztol
  nvec = np.arange(n)
  phi = -np.ones(n, dtype=int)  # null assignment
  y0 = np.zeros(n)  # null cost
  z0 = -np.ones(n, dtype=int)  # inverse assignment
  icol = np.zeros(n, dtype=bool)
  irow = 0
  for rsel, row in zip(sel, cmat):
    j = np.argmin(row[rsel])
    k = nvec[rsel][j]  # candidate
    if not icol[k]:
      phi[irow] = k
      y0[k] = row[k]
      z0[k] = irow
      icol[k] = True
    else:
      if row[k] <= y0[k]:
        phi[irow] = k
        y0[k] = row[k]
        phi[z0[k]] = -1  # unassign previous
        z0[k] = irow
    irow += 1
  return phi

def transpose_map(phi):
  n = len(phi)
  row = -np.ones(n, dtype=int)
  for i, j in enumerate(phi):
    if j >= 0:
      row[j] = i
  return row

def full_set(n):
  return set(range(n))

def minimum_reduced_cost(cbar, su, lv):
  m, n = cbar.shape
  assert m == n
  vset = full_set(n)
  delta = np.inf
  for i in su:
    for j in vset-lv:
      if cbar[i, j] < delta:
        delta = cbar[i, j]
  return delta

def alternate(k, row, cbar, verbose=False, return_path=False):
  n = len(row)
  jpred = -np.ones(n, dtype=int)
  su = set()  # selected rows
  lv = set()  # labeled columns
  sv = set()  # selected columns
  vset = full_set(n)
  i = k
  fail = False
  sink = -1
  if return_path:
    path = []
  iwhile = 0
  while (not fail) and (sink < 0):
    if verbose:
      print(iwhile, sink, fail)
    su.add(i)
    for j in vset-lv:  # look for admissible column at row i
      if (cbar[i, j] == 0):  # admissible
        lv.add(j)
        jpred[j] = i
    # scan lv by selecting one column at a time
    if len(lv-sv) > 0:  # scan next column in lv
      j = (lv-sv).pop()
      sv.add(j)
      if row[j] < 0:
        sink = j
        if return_path:
          path.append((jpred[j], j))
      else:
        #if return_path:
        #  path.append((jpred[j], j))
        if return_path:
          path.append((i, j))
        i = row[j]
        if return_path:
          path.append((i, j))
          #path.append((jpred[j], j))
    else:  # finished scan without finding sink
      fail = True
    iwhile += 1
  if verbose:
    print(iwhile, sink, fail)
  ret = (sink, jpred, su, lv, sv)
  if return_path:
    ret = (sink, jpred, su, lv, sv, path)
  return ret

def alternating_path(k, jpred, sink, row0, phi0):
  assert sink >= 0
  row = row0.copy()
  phi = phi0.copy()
  j = sink
  i = -1
  ijl = []
  while i != k:
    ijl.append(j)
    i = jpred[j]
    row[j] = i
    ijl.append(i)
    phi[i], j = j, phi[i]
  #print(sink)
  #print(ijl)
  #print(k)
  path = []
  #path.append((ijl[0], -1))
  for i, j in zip(ijl[:-1], ijl[1:]):
    #path.append((i, j))
    path.append((j, i))
  #path.append((-1, ijl[-1]))
  print('alt', path)
  return path

def cost(cmat, phi):
  sel = phi >= 0
  z = 0
  for row, j in zip(cmat[sel], phi[sel]):
    z += row[j]
  return z

def show_cost_matrix(ax, cmat):
  m, n = cmat.shape
  assert m == n
  x = np.array(n*[np.arange(n)], dtype=int).ravel()
  y = np.array([[irow]*n for irow in range(n)], dtype=int).ravel()
  ax.matshow(cmat)
  ax.get_xaxis().tick_bottom()
  ## !!!! y should start from the top
  #yticks = range(n)
  #ax.set_yticks(yticks)
  #ax.set_yticklabels(yticks[::-1])
  for x1, y1, label in zip(x, y,cmat.ravel().astype(str)): 
    ax.text(x1, y1, label, ha='center', va='center',
      c='white', fontsize=36)

def show_assignment(ax, phi, ms=36):
  sel = phi >= 0
  y = np.where(sel)[0]
  x = phi[sel]
  ax.plot(x, y, ls='', marker='o', fillstyle='none',
    ms=ms, c='r', mew=5,
  )

def show_path(ax, path):
  x, y = np.array(path).T
  ax.plot(y, x, lw=8)

def plot_solution(cmat, cbar, phi, path=None, i0=None, j0=None,
  fig_loc = None):
  fig, axl = plt.subplots(1, 2, sharex=True, sharey=True)
  axl[0].set_xlabel('cost')
  axl[1].set_xlabel('reduced cost')
  show_cost_matrix(axl[0], cmat)
  show_cost_matrix(axl[1], cbar)
  for ax in axl:
    show_assignment(ax, phi)
    if path is not None:
      axl[0].set_title('row %d with col %d' % (i0, j0))
      show_path(ax, path)
  fig.tight_layout()
  fig.subplots_adjust(hspace=0.025)
  if fig_loc is not None:
    fig.savefig(fig_loc, dpi=320)
  plt.show()

def kuhn_hungarian(cmat, mwhile=1024*1024, verbose=False):
  m, n = cmat.shape
  assert m == n
  u, v = dual_variables(cmat)
  cbar = admissible_transform(cmat, u, v)
  # use my initial
  phi = partial_primal_solution(cmat)
  row = transpose_map(phi)
  # main loop
  iwhile = 0
  uset = set(np.where(phi>=0)[0])
  while (len(uset) < n) and (iwhile < mwhile):
    k = (full_set(n)-uset).pop()
    # find alternating path
    sink, jpred, su, lv, sv, path = alternate(k, row, cbar,
      return_path=True)
    if verbose:
      print(iwhile, sink, jpred)
    if sink >= 0:  # enlarge primal solution
      uset.add(k)
      j = sink
      i = -1
      while i != k:
        i = jpred[j]
        row[j] = i
        phi[i], j = j, phi[i]
    else:  # update the dual solution
      delta = minimum_reduced_cost(cbar, su, lv)
      for i in su:
        u[i] += delta
      for j in lv:
        v[j] -= delta
      cbar = admissible_transform(cmat, u, v)
    uset = set(np.where(phi>=0)[0])
    iwhile += 1
  return phi

if __name__ == '__main__':
  outdir = sys.argv[2]
  fmat = sys.argv[1]
  cmat = np.loadtxt(fmat)

  start = timer()
  phi = kuhn_hungarian(cmat)
  z = cost(cmat, phi)
  end = timer()
  print(end-start)

  fout = os.path.join(outdir, 'out3.dat')
  np.savetxt(fout, [z])
  with open(fout, 'a') as f:
    np.savetxt(f, phi, fmt='%d')
# end __main__
